"use strict";
var _this = this;
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var chai_1 = require("chai");
var memFs = require("mem-fs");
var fsEditor = require("mem-fs-editor");
var fs_1 = require("fs");
var ejs_1 = require("ejs");
var prettier_1 = require("prettier");
var uglify_es_1 = require("uglify-es");
var immutable_1 = require("immutable");
var migrate_app_js_1 = require("../../steps/migrate_app_js");
describe("migrate app js", function () {
    var editor, originalSrc, options;
    var cwd = process.cwd();
    var src = cwd + "/src/test/fixtures/migrate_js_app";
    var dest = cwd + "/tmp/test/migrate_js_app";
    beforeEach(function () {
        editor = fsEditor.create(memFs.create());
        options = immutable_1.Map({ src: src, dest: dest, editor: editor });
    });
    describe("with an empty v1 app.js file", function () {
        beforeEach(function () {
            originalSrc = editor.read(src + "/app.js");
            editor.write(src + "/app.js", "");
        });
        afterEach(function () { return editor.write(src + "/app.js", originalSrc); });
        it("should insert no-op app code into legacy_app.js", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var code, expected;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, migrate_app_js_1.default(options)];
                    case 1:
                        _a.sent();
                        code = prettier_1.format(editor.read(dest + "/src/javascripts/legacy_app.js"));
                        expected = prettier_1.format("\n      const App = (function() {\n        return {\n          /* no-op */\n        };\n      }());\n      ");
                        chai_1.expect(code).to.have.string(expected);
                        return [2 /*return*/];
                }
            });
        }); });
    });
    describe("with a v1 app.js file", function () {
        it("should extract the v1 return statement into src/javascripts/legacy_app.js", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var code, expected;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, migrate_app_js_1.default(options)];
                    case 1:
                        _a.sent();
                        code = uglify_es_1.minify(editor.read(dest + "/src/javascripts/legacy_app.js")).code;
                        expected = uglify_es_1.minify("\n        const App = (function() {\n          return {\n            events: { \"app.activated\": \"init\" },\n            foo: require(\"./lib/foo\"),\n            init() {}\n          };\n        })()\n      ").code;
                        chai_1.expect(code).to.have.string(expected);
                        return [2 /*return*/];
                }
            });
        }); });
        it("should ignore any code outside of the v1 IIFE", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var code;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, migrate_app_js_1.default(options)];
                    case 1:
                        _a.sent();
                        code = editor.read(dest + "/src/javascripts/legacy_app.js");
                        chai_1.expect(code).not.to.match(/function someNaughtyFunction\(\)/);
                        return [2 /*return*/];
                }
            });
        }); });
        describe("when there are common js modules", function () {
            beforeEach(function () {
                options = options.set("hasCommonJs", true);
            });
            it("should update require statements in app.js to be relative", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, migrate_app_js_1.default(options)];
                        case 1:
                            _a.sent();
                            chai_1.expect(editor.read(dest + "/src/javascripts/legacy_app.js")).to.match(/require\("\.\/lib\/foo"\)/);
                            return [2 /*return*/];
                    }
                });
            }); });
        });
        describe("given the --auto flag is true", function () {
            beforeEach(function () {
                options = options.set("auto", true);
            });
            var readMigratedSrc = function () {
                var js = editor.read(dest + "/src/javascripts/legacy_app.js");
                var code = uglify_es_1.minify(js, {
                    mangle: false,
                    compress: false,
                    ecma: 6
                }).code;
                return code;
            };
            var wrapExpectedSrc = function (js, async) {
                if (async === void 0) { async = true; }
                js = wrapSrcWithExpression(js);
                var tpl = fs_1.readFileSync("src/templates/legacy_app.ejs", {
                    encoding: "utf-8"
                });
                js = ejs_1.render(tpl, { code: js, helpers: { async: async } });
                var code = uglify_es_1.minify(js, {
                    mangle: false,
                    compress: false,
                    ecma: 6
                }).code;
                return code;
            };
            var writeFixtureSrc = function (js) {
                editor.write(src + "/app.js", wrapSrcWithExpression(js));
            };
            var wrapSrcWithExpression = function (js) {
                return "(function() { return { " + js + " }; })();";
            };
            describe("with invoke", function () {
                beforeEach(function () {
                    editor.writeJSON(src + "/manifest.json", {
                        location: "nav_bar"
                    });
                });
                it("should migrate the v1 api to be async/await", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                writeFixtureSrc("foo: function() {\n            this.preloadPane();\n          }");
                                return [4 /*yield*/, migrate_app_js_1.default(options)];
                            case 1:
                                _a.sent();
                                chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n              await wrapZafClient(this.zafClient, \"preloadPane\");\n            }"));
                                return [2 /*return*/];
                        }
                    });
                }); });
                it("shouldn't create any bindings for repeated api calls", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                writeFixtureSrc("foo: function() {\n            this.preloadPane();\n            this.preloadPane();\n          }");
                                return [4 /*yield*/, migrate_app_js_1.default(options)];
                            case 1:
                                _a.sent();
                                chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n              await wrapZafClient(this.zafClient, \"preloadPane\");\n              await wrapZafClient(this.zafClient, \"preloadPane\");\n            }"));
                                return [2 /*return*/];
                        }
                    });
                }); });
            });
            ["ticket", "user", "organization"].forEach(function (location) {
                describe("when the v1 app is in the " + location + " location", function () {
                    beforeEach(function () {
                        editor.writeJSON(src + "/manifest.json", {
                            location: location + "_sidebar"
                        });
                    });
                    describe("with " + location + " APIs", function () {
                        it("should migrate v1 APIs to be async/await", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  console.log(this." + location + "().requester().email());\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                const _" + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                console.log(_" + location + ".requester.email);\n              }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should make other methods async, if they call an async method", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  var " + location + " = this." + location + "();\n                  console.log(" + location + ".requester().email());\n                },\n                bar: function() {\n                  this.foo();\n                  return true;\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                  var " + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                  console.log(" + location + ".requester.email);\n                },\n                bar: async function() {\n                  await this.foo();\n                  return true;\n                }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should migrate v1 APIs when they are in nested statements", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  return {\n                    quux: this." + location + "()\n                  };\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                  const _" + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                  return {\n                    quux: _" + location + "\n                  };\n                }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should create unique var names to avoid conflict with existing bindings", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  var " + location + " = true;\n                  var requester = this." + location + "().requester();\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                  const _" + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                  var " + location + " = true;\n                  var requester = _" + location + ".requester;\n                }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should reuse existing bindings, where possible", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  var " + location + " = this." + location + "();\n                  console.log(" + location + ".requester().email());\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                var " + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                console.log(" + location + ".requester.email);\n              }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should only care about the base API", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                  this.some.thing." + location + ".foo();\n                }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: function() {\n                this.some.thing." + location + ".foo();\n              }", false));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        it("should handle cases where bind or then are chained onto a parent expression", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        writeFixtureSrc("foo: function() {\n                this.someMethod(this." + location + "()).then(() => {});\n              }");
                                        return [4 /*yield*/, migrate_app_js_1.default(options)];
                                    case 1:
                                        _a.sent();
                                        chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                    const _" + location + " = await wrapZafClient(this.zafClient, \"" + location + "\");\n                    this.someMethod(_" + location + ").then(() => {});\n                  }"));
                                        return [2 /*return*/];
                                }
                            });
                        }); });
                        var api = location + "Fields";
                        describe(api, function () {
                            describe("with no arguments", function () {
                                it("should migrate v1 APIs to be async/await", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    return tslib_1.__generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0:
                                                writeFixtureSrc("foo: function() {\n                     const fields = this." + api + "();\n                    }");
                                                return [4 /*yield*/, migrate_app_js_1.default(options)];
                                            case 1:
                                                _a.sent();
                                                chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                      const fields = await wrapZafClient(this.zafClient, \"" + api + "\");\n                     }"));
                                                return [2 /*return*/];
                                        }
                                    });
                                }); });
                            });
                            describe("with arguments", function () {
                                it("should pass the api and field names to the wrapZafClient helper", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    return tslib_1.__generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0:
                                                writeFixtureSrc("foo: function() {\n                     const fields = this." + api + "(\"brand\");\n                    }");
                                                return [4 /*yield*/, migrate_app_js_1.default(options)];
                                            case 1:
                                                _a.sent();
                                                chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                      const fields = await wrapZafClient(this.zafClient, \"" + api + "\", \"brand\");\n                     }"));
                                                return [2 /*return*/];
                                        }
                                    });
                                }); });
                                it("should pass the api, field names, and value to set to the wrapZafClient helper", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    return tslib_1.__generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0:
                                                writeFixtureSrc("foo: function() {\n                    const fields = this." + api + "(\"brand\", \"foo\");\n                   }");
                                                return [4 /*yield*/, migrate_app_js_1.default(options)];
                                            case 1:
                                                _a.sent();
                                                chai_1.expect(readMigratedSrc()).to.have.string(wrapExpectedSrc("foo: async function() {\n                      const fields = await wrapZafClient(this.zafClient, \"" + api + "\", \"brand\", \"foo\"); \n                     }"));
                                                return [2 /*return*/];
                                        }
                                    });
                                }); });
                            });
                        });
                    });
                });
            });
        });
    });
});
